/**
 * AutoDAO - Generic DAO on steroids implementation for Java.
 * Copyright (C) 2008-2009  Marat Radchenko <slonopotamus@users.sourceforge.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package net.sf.autodao.hibernate;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import static java.util.Arrays.asList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.autodao.Dao;
import net.sf.autodao.DefaultQueryArgumentTransformer;
import net.sf.autodao.Finder;
import net.sf.autodao.FinderIntroductionInterceptor;
import net.sf.autodao.Named;
import net.sf.autodao.PersistentEntity;
import net.sf.autodao.QueryArgumentTransformer;
import net.sf.autodao.SpecificationViolationException;
import net.sf.autodao.Utils;
import org.hibernate.MappingException;
import org.hibernate.QueryException;
import org.hibernate.SessionFactory;
import org.hibernate.engine.SessionFactoryImplementor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.aop.framework.ProxyFactoryBean;
import org.springframework.beans.factory.BeanFactory;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.orm.hibernate3.HibernateTemplate;
import org.springframework.transaction.PlatformTransactionManager;
import org.springframework.transaction.TransactionDefinition;
import org.springframework.transaction.interceptor.DefaultTransactionAttribute;
import org.springframework.transaction.interceptor.MatchAlwaysTransactionAttributeSource;
import org.springframework.transaction.interceptor.TransactionInterceptor;
import org.springframework.util.Assert;
import static org.springframework.util.StringUtils.hasText;

/**
 * Validating proxy factory.
 * <p/>
 * NOT FOR PUBLIC USE.
 */
final class HibernateDaoProxyFactoryBean extends ProxyFactoryBean {
  private static final String QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME = "autodaoQueryArgumentTransformer";

  /** Log. Just log. */
  private static final Logger LOG = LoggerFactory.getLogger(HibernateDaoProxyFactoryBean.class);

  public HibernateDaoProxyFactoryBean(final Class<?> daoInterface,
                                      final SessionFactory sessionFactory,
                                      final PlatformTransactionManager transactionManager,
                                      final Class<? extends PersistentEntity<?>> entityType) {
    Assert.notNull(daoInterface, "No DAO interface");
    addInterface(daoInterface);
    Assert.notNull(sessionFactory, "No session factory set");
    validate(daoInterface, sessionFactory);
    Assert.notNull(transactionManager);
    final Class<? extends PersistentEntity<?>> realEntityType = getEntityType(daoInterface, entityType);
    addAdvice(createTransactionInterceptor(transactionManager));
    setTarget(createDao(daoInterface, realEntityType, sessionFactory));
  }

  private static TransactionInterceptor createTransactionInterceptor(final PlatformTransactionManager transactionManager) {
    final MatchAlwaysTransactionAttributeSource source = new MatchAlwaysTransactionAttributeSource();
    source.setTransactionAttribute(new DefaultTransactionAttribute(TransactionDefinition.PROPAGATION_MANDATORY));
    return new TransactionInterceptor(transactionManager, source);
  }

  @Override
  public void setBeanFactory(final BeanFactory beanFactory) {
    QueryArgumentTransformer queryArgumentTransformer = new DefaultQueryArgumentTransformer();
    try {
      queryArgumentTransformer = (QueryArgumentTransformer) beanFactory.getBean(QUERY_ARGUMENT_TRANSFORMER_BEAN_NAME, QueryArgumentTransformer.class);
    } catch (final NoSuchBeanDefinitionException ignore) {
      // Use default
    }
    addAdvice(new FinderIntroductionInterceptor(queryArgumentTransformer));
    super.setBeanFactory(beanFactory);
  }

  private static Class<? extends PersistentEntity<?>> getEntityType(final Class<?> daoInterface,
                                                                    final Class<? extends PersistentEntity<?>> entityType) {
    Class<? extends PersistentEntity<?>> realEntityType;
    if (Dao.class.isAssignableFrom(daoInterface)) {
      if (entityType == null) {
        realEntityType = autoDiscoverEntityType(daoInterface);
      } else {
        realEntityType = entityType;
      }
      Assert.notNull(realEntityType, "No entity type found for " + daoInterface);
      //See NoEntityTest
      Assert.isAssignable(PersistentEntity.class, realEntityType, "Invalid entity type: ");
    } else {
      if (entityType != null) {
        LOG.warn("Useless entity type for dao " + daoInterface + ", skipping...");
      }
      realEntityType = null;
    }
    return realEntityType;
  }

  @SuppressWarnings({ "unchecked" })
  @NotNull
  private static Object createDao(@NotNull final Class<?> daoInterface,
                                  @Nullable final Class<?> entityType,
                                  @NotNull final SessionFactory sessionFactory) {
    final HibernateTemplate hibernateTemplate = new HibernateTemplate(sessionFactory, false);
    if (Dao.class.isAssignableFrom(daoInterface)) {
      assert entityType != null;
      Assert.notNull(sessionFactory.getClassMetadata(entityType), entityType.getName()
          + " is not a persistent class (it is not registered in SessionFactory)");
      return new HibernateEntityDao(entityType, hibernateTemplate);
    } else {
      return new HibernateDao(hibernateTemplate);
    }
  }

  @SuppressWarnings({ "unchecked" })
  private static Class<? extends PersistentEntity<?>> autoDiscoverEntityType(final Class<?> daoInterface) {
    //Automagically discover entity type from generics.
    for (final Type t : daoInterface.getGenericInterfaces()) {
      if (t instanceof ParameterizedType) {
        final ParameterizedType pt = (ParameterizedType) t;
        if (pt.getRawType() == Dao.class) {
          final Type type = pt.getActualTypeArguments()[0];
          if (type instanceof Class) {
            return (Class<? extends PersistentEntity<?>>) type;
          } else if (type instanceof ParameterizedType) {
            //Special handling for cases like Dao<Entity<Foo>>.
            //See ParamEntityTest
            return (Class<? extends PersistentEntity<?>>) ((ParameterizedType) type).getRawType();
          } else {
            //TODO: is it actually possible?
            throw new IllegalArgumentException(
                "Don't know how to handle " + type + " entity type. Please update to latest AutoDAO release."
                    + " If that doesn't help, please file a bugreport to AutoDAO bugtracker. Thanks.");
          }
        }
      }
    }
    return null;
  }

  /**
   * Validates given DAO interface.
   * @param daoInterface   dao interface.
   * @param sessionFactory Hibernate SessionFactory.
   */
  private static void validate(final Class<?> daoInterface, final SessionFactory sessionFactory) {
    final SessionFactoryImplementor sfi;
    if (sessionFactory instanceof SessionFactoryImplementor) {
      sfi = (SessionFactoryImplementor) sessionFactory;
    } else {
      sfi = null;
      LOG.warn("Could not cast session factory " + sessionFactory + " to "
          + SessionFactoryImplementor.class.getName() + ". Some validations were disabled for DAO "
          + daoInterface.getName());
    }

    for (final Method m : daoInterface.getMethods()) {
      final Finder f = m.getAnnotation(Finder.class);
      if (f == null) {
        if (!asList(Dao.class.getMethods()).contains(m)) {
          throw new SpecificationViolationException("Method is neither a finder method nor is implemented in AutoDAO", m);
        }
      } else {
        final String query = f.query();
        final String queryName = f.queryName();
        if (hasText(query) && hasText(queryName)) {
          throw new SpecificationViolationException("You cannot specify query and query name at the same time", m);
        } else if (!hasText(query) && !hasText(queryName)) {
          throw new SpecificationViolationException("Neither query nor query name specified", m);
        }
        if (sfi != null) {
          if (hasText(query)) {
            try {
              sfi.getQueryPlanCache().getHQLQueryPlan(query, false, Collections.emptyMap());
            } catch (final QueryException e) {
              throw new SpecificationViolationException("Invalid query: " + query, m, e);
            } catch (final MappingException e) {
              throw new SpecificationViolationException("Invalid query" + query, m, e);
            }
            LOG.info("Validated query '" + query + "' for method " + m);
          } else if (hasText(queryName) && sfi.getNamedQuery(queryName) == null) {
            throw new SpecificationViolationException("No named query '" + queryName + "' found.", m);
          }
        }
        if (Collection.class.isAssignableFrom(m.getReturnType())) {
          final Class<?> returnType = m.getReturnType();
          if (!returnType.isAssignableFrom(f.returnAs())) {
            throw new SpecificationViolationException(
                "@Finder.returnAs value (" + f.returnAs().getName()
                    + ") cannot be cast to method return type (" + returnType.getName() + ")", m
            );
          } else if (f.returnAs() != List.class) {
            if (f.returnAs().isInterface()) {
              throw new SpecificationViolationException(
                  "@Finder.returnAs value (" + f.returnAs().getName() + ") cannot be an interface", m
              );
            } else {
              try {
                f.returnAs().getConstructor();
              } catch (NoSuchMethodException e) {
                throw new SpecificationViolationException(
                    "@Finder.returnAs value (" + f.returnAs().getName() + ") doesn't have no-arg constructor.", m
                );
              }
            }
          }
        }
        validateParams(m);
      }
    }
  }

  private static void validateParams(final Method meth) {
    final boolean isSingleFind = !Collection.class.isAssignableFrom(meth.getReturnType());
    final Integer limitIndex = Utils.getLimitParameterIndex(meth);
    final Integer offsetIndex = Utils.getOffsetParameterIndex(meth);
    final Set<String> names = new HashSet<String>();
    Boolean hasNamed = null;
    for (int i = 0; i < meth.getParameterAnnotations().length; i++) {
      final Class<?> paramType = meth.getParameterTypes()[i];
      final Named named = Utils.getNamed(meth.getParameterAnnotations()[i]);
      final boolean isOffset = Utils.getOffset(meth.getParameterAnnotations()[i]) != null;
      final boolean isLimit = Utils.getLimit(meth.getParameterAnnotations()[i]) != null;
      if (!isLimit && !isOffset) {
        if (hasNamed == null) {
          hasNamed = named != null;
        } else if (hasNamed && named == null || !hasNamed && named != null) {
          throw new SpecificationViolationException("You're not allowed to mix @Named and indexed params", meth);
        }
      } else if (named != null) {
        throw new SpecificationViolationException("@Named cannot be present on @Limit/@Offset param", meth);
      }
      if (named != null && !names.add(named.value())) {
        throw new SpecificationViolationException("Duplicate named parameters '" + named.value() + "'", meth);
      }
      if (isSingleFind) {
        if (isLimit || isOffset) {
          throw new SpecificationViolationException("Neither @Limit nor @Offset is allowed on single-object finders",
              meth);
        }
      } else {
        if ((isLimit || isOffset) && !Integer.class.isAssignableFrom(paramType)
            && !Integer.TYPE.isAssignableFrom(paramType)) {
          throw new SpecificationViolationException(
              "@Limit and @Offset argument type must be either int or java.lang.Integer", meth);
        }
        if (isLimit && limitIndex != i || isOffset && offsetIndex != i) {
          throw new SpecificationViolationException("@Limit and @Offset can appear only on one parameter", meth);
        }
      }
    }
  }
}
